#include <string.h>
#include <gtk/gtk.h>
#include <expat.h>
#include <glib/gi18n.h>
#include "scp_log.h"
#include "scp_session.h"
#include "scp_config.h"

#define SCPLAYER_SYSTEM_TRAY_ICON   "/usr/share/icons/Faenza/apps/scalable/totem.svg"
#define SCPLAYER_SYSTEM_TRAY_ICON_NAME  "totem"
#define CH_XML_READ_BUF_SIZE    (512*1024)

enum {SCP_CH_RLD_IDLE, SCP_CH_RLD_START, SCP_CH_RLD_PARSE, SCP_CH_RLD_END, SCP_CH_RLD_ERROR};

typedef struct scp_menu ScpMenu;

typedef struct sc_channel {
	ScpMenu* scplayer;
	GString* name;
	GString* url;
	GtkMenuItem* menuItem;
} SCChannel;

struct scp_menu{
    XML_Parser xmlParser;
    GFile* chFile;
    GFileInputStream* chIS;
    GBufferedInputStream* chBIS;
    GtkMenu* currGrpM;
    GList* channels;
    int chReloadStatus;
    const char* uri;
    GtkStatusIcon* icon;
    GtkMenu* menu;
    GtkMenuItem* quitMI;
    GtkMenuItem* aboutMI;
    GtkMenuItem* settingsMI;
    GtkMenuItem* refreshMI;
    GtkMenuItem* chGroupMI;
    GtkMenu* groupMenu;
    GtkDialog* buferringDlg;
    ScpConfig config;
};

int ScpAsyncReloadChannels(ScpMenu* this);
int ScpMenuReloadChannels(ScpMenu* this);


int SCChannelInit(SCChannel* scc, const char* name, const char* url, ScpMenu* scplayer)
{
	scc->scplayer = scplayer;
	scc->name = g_string_new(name);
	scc->url = g_string_new(url);
	scc->menuItem = (GtkMenuItem*)gtk_menu_item_new_with_label(name);
	gtk_widget_show((GtkWidget*)scc->menuItem);
	return 0;
}

SCChannel* SCChannelNew(const char* name, const char* url, ScpMenu* scplayer)
{
	SCChannel* ret;
	
	ret = (SCChannel*)malloc(sizeof(SCChannel));
	if (SCChannelInit(ret, name, url, scplayer)) {
		return NULL;
	}
	return ret;
}

int SCChannelExit(SCChannel* scc)
{
	g_object_unref(scc->menuItem);
	g_string_free(scc->name, TRUE);
	g_string_free(scc->url, TRUE);
	scc->scplayer = NULL;
	return 0;
}

int SCChannelDestroy(SCChannel* scc)
{
	SCChannelExit(scc);
	free(scc);
	return 0;
}

int SCChannelSetName(SCChannel* scc, const char* name)
{
	g_string_assign(scc->name, name);
	gtk_menu_item_set_label(scc->menuItem, name);
	return 0;
}

int SCChannelSetUrl(SCChannel* scc, const char* url)
{
	g_string_assign(scc->url, url);
	return 0;
}

gboolean ScpMenuIconOnclick(GtkStatusIcon *this, GdkEvent *event, gpointer user_data)
{
    ScpMenu* sm = (ScpMenu*)user_data;
    gtk_menu_popup(sm->menu, NULL, NULL, gtk_status_icon_position_menu, this, 0, gtk_get_current_event_time());
    return TRUE;
}

void ScpMenuQuitMIOnclick(GtkMenuItem* item, gpointer userData)
{
    gtk_main_quit();
}

void ScpMenuSettingsMIOnclick(GtkMenuItem* item, gpointer userData)
{
   ScpMenu* sm;
   
   sm = (ScpMenu*)userData;
   ScpConfigShowDialog(&sm->config);
}

/* Function to open a dialog box displaying the message provided. */
void ScpMenuAboutMIOnclick(GtkMenuItem* item, gpointer user_data) 
{
	const char* authors[] = {_("Chunfeng Zhang<CrazyPandar@gmail.com>"), NULL}; 
	gtk_show_about_dialog (NULL,
                       "program-name", "SCPlayer",
                       "logo-icon-name", "totem",
                       "title", _("About SCPlayer"),
                       "authors", authors,
                       "comments", _("SCPlayer is a GUI frontend for sopcast."),
                       "copyright", _("Copyright © 2011 Chunfeng Zhang"),
                       "website", "http://code.google.com/p/scplayer",
                       "website-label", _("SCPlayer website"),
                       NULL);
}

void ScpMenuRefreshOnclick(GtkMenuItem* item, gpointer userData)
{
	ScpMenu* sm;
	
	sm = (ScpMenu*)userData;
	ScpMenuReloadChannels(sm);
}

enum{
    ElemRoot = 0,
    ElemChannels, 
    ElemChannelsGroup, 
    ElemChannelsGroupChannel, 
    ElemChannelsGroupChannelName,
    ElemChannelsGroupChannelSopAddr, 
    ElemChannelsGroupChannelSopAddrItem
};
    
static int CurElem = ElemRoot;
static GtkMenuItem *CurGrpMI = NULL;
static GtkMenu *CurChMenu = NULL;
static SCChannel* CurCh;
void ChXmlElemSartHnd(void* usrData, const XML_Char *name, const XML_Char **atts)
{
    int i;
    ScpMenu *sm = (ScpMenu*)usrData;
    const char *langTag;
    
    if(!(langTag = getenv("LC_ALL")) && !(langTag = getenv("LC_MESSAGE"))){
        langTag = "en";
    }else{
        if(strstr(langTag, "zh_CN")){
            langTag = "cn";
        }else{
            langTag = "en";
        }
    }
    
    if (!strcmp(name, "item")) {
        switch (CurElem) {
            case ElemChannelsGroupChannelSopAddr:
                CurElem = ElemChannelsGroupChannelSopAddrItem;
                break;
        }
    } else if (!strcmp(name, "name")) {
        switch (CurElem) {
            case ElemChannelsGroupChannel:
                CurElem = ElemChannelsGroupChannelName;
                    for(i=0; atts[i]; i+=2){
                        if(!strcmp(atts[i], langTag)){
                            SCChannelSetName(CurCh, atts[i+1]);
                            break;
                        }
                    }
                break;
        }
    } else if (!strcmp(name, "sop_address")) {
        switch (CurElem) {
            case ElemChannelsGroupChannel:
                CurElem = ElemChannelsGroupChannelSopAddr;
                break;
        }
    } else if (!strcmp(name, "channel")) {
        switch (CurElem) {
            case ElemChannelsGroup:
                CurElem = ElemChannelsGroupChannel;
                CurCh = SCChannelNew("", "", sm);
                break;
        }
    } else if (!strcmp(name, "group")) {
        switch (CurElem) {
            case ElemChannels:
                CurElem = ElemChannelsGroup;
                    for(i=0; atts[i]; i+=2){
                        if(!strcmp(atts[i], langTag)){
                            SCP_DBG("Find group: %s\n", atts[i+1]);
                            if (!atts[i+1][0]) {
                                //CurGrpMI = (GtkMenuItem*)gtk_menu_item_new_with_label("unknown");
                                SCP_DBG("ignoring unknown name group\n");
                                break;
                    		} else {
                    		    CurGrpMI = (GtkMenuItem*)gtk_menu_item_new_with_label(_(atts[i+1]));
                    		}
                    		gtk_widget_show((GtkWidget*)CurGrpMI);
                    		gtk_menu_shell_append((GtkMenuShell*)sm->groupMenu, (GtkWidget*)CurGrpMI);
                    		CurChMenu = (GtkMenu*)gtk_menu_new();
                    		gtk_widget_show((GtkWidget*)CurChMenu);
                    		gtk_menu_item_set_submenu((GtkMenuItem*)CurGrpMI, (GtkWidget*)CurChMenu);
                    		break;
                        }
                    }
                break;
        }
    }else if (!strcmp(name, "channels")) {
        switch (CurElem) {
            case ElemRoot:
                CurElem = ElemChannels;
                break;
        }
    }
}

void ChXmlElemEndHnd(void *userData, const XML_Char *name)
{
    if (!strcmp(name, "item")) {
        switch (CurElem) {
            case ElemChannelsGroupChannelSopAddrItem:
                CurElem = ElemChannelsGroupChannelSopAddr;
                break;
        }
    } else if (!strcmp(name, "name")) {
        switch (CurElem) {
            case ElemChannelsGroupChannelName:
                CurElem = ElemChannelsGroupChannel;
                break;
        }
    } else if (!strcmp(name, "sop_address")) {
        switch (CurElem) {
            case ElemChannelsGroupChannelSopAddr:
                CurElem = ElemChannelsGroupChannel;
                break;
        }
    } else if (!strcmp(name, "channel")) {
        switch (CurElem) {
            case ElemChannelsGroupChannel:
                CurElem = ElemChannelsGroup;
                break;
        }
    } else if (!strcmp(name, "group")) {
        switch (CurElem) {
            case ElemChannelsGroup:
                CurElem = ElemChannels;
                break;
        }
    }else if (!strcmp(name, "channels")) {
        switch (CurElem) {
            case ElemChannels:
                CurElem = ElemRoot;
                break;
        }
    }
}

static ScpSession session;

void _ChanelMIOnclick(GtkMenuItem *menuitem, gpointer data)
{
	SCChannel* ch;
	static int firstRun = 1;
	
	ch = (SCChannel*)data;
	SCP_DBG("Playing %s, %s\n", ch->name->str, ch->url->str);
	if (firstRun) {
		firstRun = 0;
	} else {
		ScpSessionStop(&session);
	}
	ScpSessionRun(&session, ch->scplayer->config.scExe->str, 3908, 8908, ch->scplayer->config.playerExe->str, ch->url->str);
}

void chDataHnd(void *userData, const XML_Char *s, int len)
{
    if (CurElem == ElemChannelsGroupChannelSopAddrItem) {
    	GString *tmpStr = g_string_new_len(s, len);
        SCChannelSetUrl(CurCh, tmpStr->str);
        g_string_free(tmpStr, TRUE);
        g_signal_connect(CurCh->menuItem, "activate", G_CALLBACK(_ChanelMIOnclick), CurCh);
        g_object_ref((GObject*)CurCh->menuItem);
        gtk_menu_shell_append((GtkMenuShell*)CurChMenu, (GtkWidget*)CurCh->menuItem);
    }
}

void ScpChReadable(GObject* obj, GAsyncResult* res, gpointer usrData)
{
    ScpMenu *sm = (ScpMenu*)usrData;
    GBufferedInputStream *bis = (GBufferedInputStream*)obj;
    GError *err = NULL;
    int readed, avalaible, bufSz;
    gsize dummySz;
        
    readed = g_buffered_input_stream_fill_finish(bis, res, &err);
    avalaible = g_buffered_input_stream_get_available(bis);
    bufSz = g_buffered_input_stream_get_buffer_size(bis);
    SCP_DBG("readed=%d\n", readed);
    if(readed == -1){
        sm->chReloadStatus = SCP_CH_RLD_ERROR;
    }else if(readed == 0 && avalaible < bufSz){
        sm->chReloadStatus = SCP_CH_RLD_END;
    }else if(readed ==0 && avalaible == bufSz){
        g_buffered_input_stream_set_buffer_size(bis, bufSz+CH_XML_READ_BUF_SIZE);
    }else{
        XML_Parse(sm->xmlParser, g_buffered_input_stream_peek_buffer(bis, &dummySz)+(avalaible-readed), readed, 0);
    }
    if(err){
        g_error_free(err);
    }
    ScpAsyncReloadChannels(sm);
}

int ScpAsyncReloadChannels(ScpMenu* this)
{
    switch(this->chReloadStatus){
    case SCP_CH_RLD_START:
        this->xmlParser = XML_ParserCreate(NULL);
        XML_SetUserData(this->xmlParser, this);
        XML_SetElementHandler(this->xmlParser, ChXmlElemSartHnd, ChXmlElemEndHnd);
        XML_SetCharacterDataHandler(this->xmlParser, chDataHnd);
        this->chFile = g_file_new_for_uri(this->config.chListUri->str);
        this->chIS = g_file_read(this->chFile, NULL, NULL);
        this->chBIS = (GBufferedInputStream*)g_buffered_input_stream_new((GInputStream*)this->chIS);
        g_buffered_input_stream_set_buffer_size(this->chBIS, CH_XML_READ_BUF_SIZE);
        this->chReloadStatus = SCP_CH_RLD_PARSE;
    case SCP_CH_RLD_PARSE:
        SCP_DBG("fill\n");
        g_buffered_input_stream_fill_async(this->chBIS, -1, G_PRIORITY_DEFAULT, NULL, ScpChReadable, this);
        break;
    case SCP_CH_RLD_END:
        g_object_unref(this->chBIS);
        g_object_unref(this->chIS);
        g_object_unref(this->chFile);
        XML_ParserFree(this->xmlParser);
        this->chReloadStatus = SCP_CH_RLD_IDLE;
        break;
    }
    return 0;
}

void _FreeSCChannel(void* data, void* userData)
{
	SCChannelDestroy((SCChannel*)data);
}

int ScpMenuReloadChannels(ScpMenu* this)
{
	if (this->chReloadStatus != SCP_CH_RLD_IDLE)
    	return 0;	
	g_list_foreach(this->channels, _FreeSCChannel, NULL);
    if(this->groupMenu){
        gtk_widget_destroy((GtkWidget*)this->groupMenu);
    }
    this->groupMenu = (GtkMenu*)gtk_menu_new();
    gtk_widget_show((GtkWidget*)this->groupMenu);
    this->chReloadStatus = SCP_CH_RLD_START;
    ScpAsyncReloadChannels(this);
    gtk_menu_item_set_submenu((GtkMenuItem*)this->chGroupMI, (GtkWidget*)this->groupMenu);
    return 0;
}

int ScpMenuInit(ScpMenu* this, const char* cfgFile)
{	
	int ret = -1;
	
	if ((ret = ScpConfigInit(&this->config))) {
		SCP_ERR_OUT(ERR_OUT, "Init config failed\n");
	}
    this->icon = gtk_status_icon_new_from_icon_name(SCPLAYER_SYSTEM_TRAY_ICON_NAME);
    this->menu = (GtkMenu*)gtk_menu_new();
    gtk_widget_show((GtkWidget*)this->menu);
    this->quitMI = (GtkMenuItem*)gtk_menu_item_new_with_label(_("Quit"));
    gtk_widget_show((GtkWidget*)this->quitMI);
    gtk_menu_shell_prepend((GtkMenuShell*)this->menu, (GtkWidget*)this->quitMI);
    this->aboutMI = (GtkMenuItem*)gtk_menu_item_new_with_label(_("About"));
    gtk_widget_show((GtkWidget*)this->aboutMI);
    gtk_menu_shell_prepend((GtkMenuShell*)this->menu, (GtkWidget*)this->aboutMI);
    this->settingsMI = (GtkMenuItem*)gtk_menu_item_new_with_label(_("Settins"));
    gtk_widget_show((GtkWidget*)this->settingsMI);
    gtk_menu_shell_prepend((GtkMenuShell*)this->menu, (GtkWidget*)this->settingsMI);
    this->refreshMI = (GtkMenuItem*)gtk_menu_item_new_with_label(_("Refresh Channel List"));
    gtk_widget_show((GtkWidget*)this->refreshMI);
    gtk_menu_shell_prepend((GtkMenuShell*)this->menu, (GtkWidget*)this->refreshMI);
    this->chGroupMI = (GtkMenuItem*)gtk_menu_item_new_with_label("Channels");
    gtk_widget_show((GtkWidget*)this->chGroupMI);
    gtk_menu_shell_prepend((GtkMenuShell*)this->menu, (GtkWidget*)this->chGroupMI);
    this->groupMenu = NULL;
    this->channels = NULL;
    this->chReloadStatus = SCP_CH_RLD_IDLE;
    if(ScpMenuReloadChannels(this)){
        gtk_widget_destroy((GtkWidget*)this->menu);
        SCP_ERR_OUT(ERR_CONFIG_EXIT, "Load channels list failed\n");
    }
    g_signal_connect(this->quitMI, "activate", G_CALLBACK(ScpMenuQuitMIOnclick), this);
    g_signal_connect(this->aboutMI, "activate", G_CALLBACK(ScpMenuAboutMIOnclick), this);
    g_signal_connect(this->settingsMI, "activate", G_CALLBACK(ScpMenuSettingsMIOnclick), this);
    g_signal_connect(this->refreshMI, "activate", G_CALLBACK(ScpMenuRefreshOnclick), this);
    g_signal_connect(this->icon, "button-release-event", G_CALLBACK(ScpMenuIconOnclick), this);
    return 0;
ERR_CONFIG_EXIT:
	ScpConfigExit(&this->config);
ERR_OUT:
	return ret;
}

int ScpMenuExit(ScpMenu* this)
{
    gtk_widget_destroy((GtkWidget*)this->menu);
    g_list_foreach(this->channels, _FreeSCChannel, NULL);
    ScpConfigExit(&this->config);
    return 0;
}

int main(int argc, char** argv)
{
    static ScpMenu sm;

    gtk_init(&argc, &argv);  
    if (ScpMenuInit(&sm, NULL)) {
        return -1;
    }
    gtk_main();
    ScpMenuExit(&sm);
    return 0;
}

